import json
import mimetypes
import os
import re
from builtins import RuntimeError
from tempfile import SpooledTemporaryFile
from typing import Any, Dict, Union

INMEMORY_FILESIZE = 1024 * 1024
LOCAL_STORAGE_DRIVER_NAME = "Local Storage"


def get_metadata_file_obj(metadata: Dict[str, Any]) -> "SpooledTemporaryFile[bytes]":
    f = SpooledTemporaryFile(INMEMORY_FILESIZE)
    f.write(json.dumps(metadata).encode())
    f.seek(0)
    return f


def get_content_from_file_obj(fileobj):  # type: ignore
    """Provides a real file object from file content
    Converts ``str`` and ``bytes`` to an actual file.
    """
    if isinstance(fileobj, (str, bytes)):
        f = SpooledTemporaryFile(INMEMORY_FILESIZE)
        f.write(fileobj.encode() if isinstance(fileobj, str) else fileobj)
        f.seek(0)
        return f
    elif getattr(fileobj, "file", None) is not None:
        return fileobj.file
    return fileobj


def get_filename_from_fileob(fileobj):  # type: ignore
    if getattr(fileobj, "filename", None) is not None:
        return fileobj.filename
    elif getattr(fileobj, "name", None) is not None:
        return os.path.basename(fileobj.name)
    return "unnamed"


def get_content_type_from_fileobj(fileobj, filename):  # type: ignore
    if getattr(fileobj, "content_type", None) is not None:
        return fileobj.content_type
    return mimetypes.guess_type(filename, strict=False)[0] or "application/octet-stream"


def get_content_size_from_fileobj(file):  # type: ignore
    if hasattr(file, "size"):
        return file.size
    if hasattr(file, "name"):
        try:
            return os.path.getsize(file.name)
        except (OSError, TypeError):
            pass
    if hasattr(file, "tell") and hasattr(file, "seek"):
        pos = file.tell()
        file.seek(0, os.SEEK_END)
        size = file.tell()
        file.seek(pos)
        return size
    raise RuntimeError("Unable to determine the file's size.")  # pragma: no cover


def convert_size(size: Union[str, int]) -> int:
    # convert size to number of bytes ex: 1k -> 1000; 1Ki->1024
    if isinstance(size, int):
        return size
    elif isinstance(size, str):
        pattern = re.compile(r"^(\d+)\s*(k|([KM]i?))$")
        m = re.fullmatch(pattern, size)
        if m is None:
            raise ValueError("Invalid size %s" % size)
        value, si, _ = m.groups()
        si_map = {"k": 1000, "K": 1000, "M": 1000**2, "Ki": 1024, "Mi": 1024**2}
        return int(value) * si_map[si]



import json
import tempfile
from typing import Optional

from libcloud.storage.base import Object
from libcloud.storage.types import ObjectDoesNotExistError
from sqlalchemy_file.helpers import LOCAL_STORAGE_DRIVER_NAME


class StoredFile:
    def __init__(self, obj: Object) -> None:
        if obj.driver.name == LOCAL_STORAGE_DRIVER_NAME:
            """Retrieve metadata from associated metadata file"""
            try:
                metadata_obj = obj.container.get_object(f"{obj.name}.metadata.json")
                obj.meta_data = json.load(open(metadata_obj.get_cdn_url()))
            except ObjectDoesNotExistError:
                pass
        self.name = obj.name
        self.filename = obj.meta_data.get("filename", "unnamed")
        self.content_type = obj.extra.get(
            "content_type",
            obj.meta_data.get("content_type", "application/octet-stream"),
        )
        self.object = obj

    def get_cdn_url(self) -> Optional[str]:
        try:
            return self.object.get_cdn_url()
        except NotImplementedError:
            return None

    def read(self, n: int = -1) -> bytes:
        if self.object.driver.name == LOCAL_STORAGE_DRIVER_NAME:
            return open(self.object.get_cdn_url(), "rb").read(n)
        _file = tempfile.NamedTemporaryFile()
        self.object.download(_file.name, overwrite_existing=True)
        return _file.read(n)



class ValidationError(Exception):
    """Base class for ValidationError
    Parameters:
        key: Current Column key
        msg: Validation error message
    """

    def __init__(self, key: str, msg: str):  # pragma: no cover
        super().__init__("%s: %s" % (key, msg))
        self.key = key
        self.msg = msg


class SizeValidationError(ValidationError):
    pass


class ContentTypeValidationError(ValidationError):
    pass


class InvalidImageError(ValidationError):
    pass


class DimensionValidationError(ValidationError):
    pass


class AspectRatioValidationError(ValidationError):
    pass



import uuid
from datetime import datetime
from typing import Any, Dict, List, Optional

from sqlalchemy_file.base import BaseFile
from sqlalchemy_file.helpers import (
    get_content_from_file_obj,
    get_content_size_from_fileobj,
    get_content_type_from_fileobj,
    get_filename_from_fileob,
)
from sqlalchemy_file.processors import Processor
from sqlalchemy_file.storage import StorageManager
from sqlalchemy_file.stored_file import StoredFile
from sqlalchemy_file.validators import Validator


class File(BaseFile):
    """
    Takes a file as content and uploads it to the appropriate storage
    according to the attached Column and file information into the
    database as JSON.

    Default attributes provided for all ``File`` include:

    Attributes:

        filename (str):  This is the name of the uploaded file
        file_id:   This is the generated UUID for the uploaded file
        path:            This is a  compination of `upload_storage` and `file_id` separated by
                        `/`. This will be use later to retrieve the file
        content_type:   This is the content type of the uploaded file
        uploaded_at (datetime):    This is the upload date in ISO format
        url (str):            CDN url of the uploaded file
        file:           Only available for saved content, internally call
                      [StorageManager.get_file()][sqlalchemy_file.storage.StorageManager.get_file]
                      on path and return an instance of `StoredFile`
    """

    def __init__(
        self,
        content: Any,
        filename: Optional[str] = None,
        content_type: Optional[str] = None,
    ) -> None:
        super().__init__()
        if isinstance(content, dict):
            object.__setattr__(self, "original_content", None)
            object.__setattr__(self, "saved", True)
            self.update(content)
            self._freeze()
        else:
            self.original_content = get_content_from_file_obj(content)
            filename = filename or get_filename_from_fileob(content)
            content_type = content_type or get_content_type_from_fileobj(
                content, filename
            )
            self.update(
                {
                    "filename": filename,
                    "content_type": content_type,
                    "size": get_content_size_from_fileobj(self.original_content),
                    "files": [],
                }
            )
            self._thaw()

    def apply_validators(self, validators: List[Validator], key: str = "") -> None:
        """Apply validators to current file"""
        for validator in validators:
            validator.process(self, key)

    def apply_processors(
        self,
        processors: List[Processor],
        upload_storage: Optional[str] = None,
    ) -> None:
        """Apply processors to current file"""
        for processor in processors:
            processor.process(self, upload_storage)
        self._freeze()

    def save_to_storage(self, upload_storage: Optional[str] = None) -> None:
        """Save current file into provided `upload_storage`"""
        stored_file = self.store_content(
            self.original_content,
            upload_storage,
            metadata={"filename": self.filename, "content_type": self.content_type},
        )
        self["file_id"] = stored_file.name
        self["uploaded_at"] = datetime.utcnow().isoformat()
        self["path"] = "%s/%s" % (upload_storage, stored_file.name)
        self["url"] = stored_file.get_cdn_url()
        self["saved"] = True

    def store_content(
        self,
        content: Any,
        upload_storage: Optional[str] = None,
        name: Optional[str] = None,
        metadata: Optional[Dict[str, Any]] = None,
    ) -> StoredFile:
        """Store content into provided `upload_storage`
        with additional `metadata`. Can be use by processors
        to store additional files.
        """
        name = name or str(uuid.uuid4())
        stored_file = StorageManager.save_file(name, content, upload_storage, metadata)
        self["files"].append("%s/%s" % (upload_storage, name))
        return stored_file

    def encode(self) -> Dict[str, Any]:
        return {k: v for k, v in self.items() if k not in ["original_content"]}

    @classmethod
    def decode(cls, data: Any) -> "File":
        return cls(data)

    @property
    def file(self) -> "StoredFile":
        if self.get("saved", False):
            return StorageManager.get_file(self["path"])
        raise RuntimeError("Only available for saved file")



import typing
from typing import Any, List, Tuple, TypeVar, no_type_check

from sqlalchemy.ext.mutable import Mutable

T = TypeVar("T", bound=Any)


class MutableList(Mutable, typing.List[T]):
    """
    A list type that implements :class:`Mutable`.

    The :class:`MutableList` object implements a list that will
    emit change events to the underlying mapping when the contents of
    the list are altered, including when values are added or removed.

    This is a replication of default Mutablelist provide by SQLAlchemy.
    The difference here is the properties _removed which keep every element
    removed from the list in order to be able to delete them after commit
    and keep them when session rolled back.

    """

    def __init__(self, *args, **kwargs) -> None:  # type: ignore
        super(MutableList, self).__init__(*args, **kwargs)
        self._removed: List[T] = []
        # logging.warning(('init', self._removed, args, kwargs))

    @classmethod
    def coerce(cls, key: Any, value: Any) -> Any:
        if not isinstance(value, MutableList):
            if isinstance(value, list):
                return MutableList(value)
            # this call will raise ValueError
            return Mutable.coerce(key, value)  # pragma: no cover
        else:
            return value  # pragma: no cover

    @no_type_check
    def __reduce_ex__(self, proto):  # pragma: no cover
        return self.__class__, (list(self),)

    # needed for backwards compatibility with
    # older pickles
    def __getstate__(self) -> Tuple[List[T], List[T]]:  # pragma: no cover
        return list(self), self._removed

    def __setstate__(self, state: Any) -> None:  # pragma: no cover
        self[:] = state[0]
        self._removed = state[1]

    def __setitem__(self, index: Any, value: Any) -> None:
        """Detect list set events and emit change events."""
        if isinstance(index, slice):
            old_value = self[index]
        else:
            old_value = [self[index]]
        list.__setitem__(self, index, value)
        self.changed()
        self._removed.extend(old_value)

    def __delitem__(self, index: Any) -> None:
        """Detect list del events and emit change events."""
        if isinstance(index, slice):
            old_value = self[index]
        else:
            old_value = [self[index]]
        list.__delitem__(self, index)
        self.changed()
        self._removed.extend(old_value)

    def pop(self, *arg) -> "T":  # type: ignore
        result = list.pop(self, *arg)
        self.changed()
        self._removed.append(result)
        return result

    def append(self, x: Any) -> None:
        list.append(self, x)
        self.changed()

    def extend(self, x: Any) -> None:
        list.extend(self, x)
        self.changed()

    @no_type_check
    def __iadd__(self, x):
        self.extend(x)
        return self

    def insert(self, i: Any, x: Any) -> None:
        list.insert(self, i, x)
        self.changed()

    def remove(self, i: "T") -> None:
        list.remove(self, i)
        self._removed.append(i)
        self.changed()

    def clear(self) -> None:
        self._removed.extend(self)
        list.clear(self)  # type: ignore
        self.changed()

    def sort(self, **kw: Any) -> None:
        list.sort(self, **kw)
        self.changed()

    def reverse(self) -> None:
        list.reverse(self)  # type: ignore
        self.changed()



from abc import abstractmethod
from typing import TYPE_CHECKING, List, Optional, Tuple, Union

from sqlalchemy_file.exceptions import (
    AspectRatioValidationError,
    ContentTypeValidationError,
    DimensionValidationError,
    InvalidImageError,
    SizeValidationError,
)
from sqlalchemy_file.helpers import convert_size

if TYPE_CHECKING:
    from sqlalchemy_file.file import File


class Validator:
    """
    Interface that must be implemented by file validators.
    File validators get executed before a file is stored on the database
    using one of the supported fields. Can be used to add additional data
    to file object or change it.

    """

    @abstractmethod
    def process(self, file: "File", attr_key: str) -> None:  # pragma: no cover
        """
        Should be overridden in inherited class

        Parameters:
           file: [File][sqlalchemy_file.file.File] object
           attr_key: current SQLAlchemy column key. Can be passed to
               [ValidationError][sqlalchemy_file.exceptions.ValidationError]
        """
        pass


class SizeValidator(Validator):
    """Validate file maximum size

    Attributes:
        max_size:
            If set, the size of the underlying file must
            be below this file size in order to be valid.
            The size of the file can be given in one of
            the following formats:

            | **Suffix** | **Unit Name** | **Value**       | **Example** |
            |------------|---------------|-----------------|-------------|
            | (none)     | byte          | 1 byte          | `4096`      |
            | k          | kilobyte      | 1,000 bytes     | `200k`      |
            | M          | megabyte      | 1,000,000 bytes | `2M`        |
            | Ki         | kibibyte      | 1,024 bytes     | `32Ki`      |
            | Mi         | mebibyte      | 1,048,576 bytes | `8Mi`       |

            For more information, view
            [Wikipedia: Binary prefix](https://en.wikipedia.org/wiki/Binary_prefix)
    Example:
        ```Python
        class Attachment(Base):
            __tablename__ = "attachment"

            id = Column(Integer, autoincrement=True, primary_key=True)
            name = Column(String(50), unique=True)
            content = Column(FileField(validators=[SizeValidator(max_size="2M")]))
        ```

    Raises:
        SizeValidationError: When file `size` is greater than max_size

    """

    def __init__(self, max_size: Union[int, str] = 0) -> None:
        super().__init__()
        self.max_size = max_size

    def process(self, file: "File", attr_key: str) -> None:
        if file.size > convert_size(self.max_size):
            raise SizeValidationError(
                attr_key,
                "The file is too large (%s bytes). Allowed maximum size is %s."
                % (file.size, self.max_size),
            )


class ContentTypeValidator(Validator):
    """Validate file mimetype
    Attributes:
        allowed_content_types: If set, file `content_type`
                        must be one of the provided list.

    Example:
        ```Python
        class Attachment(Base):
            __tablename__ = "attachment"

            id = Column(Integer, autoincrement=True, primary_key=True)
            name = Column(String(50), unique=True)
            content = Column(
                FileField(validators=[ContentTypeValidator(["text/plain", "text/csv"])])
            )
        ```

    Raises:
        ContentTypeValidationError: When file `content_type` not in allowed_content_types


    """

    def __init__(self, allowed_content_types: Optional[List[str]] = None) -> None:
        super().__init__()
        self.allowed_content_types = allowed_content_types

    def process(self, file: "File", attr_key: str) -> None:
        if (
            self.allowed_content_types is not None
            and file.content_type not in self.allowed_content_types
        ):
            raise ContentTypeValidationError(
                attr_key,
                "File content_type %s is not allowed. Allowed content_types are: %s"
                % (file.content_type, self.allowed_content_types),
            )


class ImageValidator(ContentTypeValidator):
    """Default Validator for ImageField

    Attributes:
        min_wh: Minimum allowed dimension (w, h).
        max_wh: Maximum allowed dimension (w, h).
        allowed_content_types: An iterable whose items are
                    allowed content types. Default is `image/*`
        min_aspect_ratio: Minimum allowed image aspect ratio.
        max_aspect_ratio: Maximum allowed image aspect ratio.

    Example:
        ```Python

        class Book(Base):
            __tablename__ = "book"

            id = Column(Integer, autoincrement=True, primary_key=True)
            title = Column(String(100), unique=True)
            cover = Column(
                ImageField(
                    image_validator=ImageValidator(
                        allowed_content_types=["image/x-icon", "image/tiff", "image/jpeg"],
                        min_wh=(200, 200),
                        max_wh=(400, 400),
                        min_aspect_ratio=1,
                        max_aspect_ratio=16/9,
                    )
                )
            )
        ```

    Raises:
        ContentTypeValidationError: When file `content_type` not in allowed_content_types
        InvalidImageError: When file is not a valid image
        DimensionValidationError: When image width and height constraints fail.

    Will add `width` and `height` properties to the file object
    """

    def __init__(
        self,
        min_wh: Optional[Tuple[int, int]] = None,
        max_wh: Optional[Tuple[int, int]] = None,
        min_aspect_ratio: Optional[float] = None,
        max_aspect_ratio: Optional[float] = None,
        allowed_content_types: Optional[List[str]] = None,
    ):
        from PIL import Image

        Image.init()
        super().__init__(
            allowed_content_types
            if allowed_content_types is not None
            else [type for type in Image.MIME.values()]
        )
        self.min_width, self.min_height = min_wh if min_wh else (None, None)
        self.max_width, self.max_height = max_wh if max_wh else (None, None)
        self.min_aspect_ratio = min_aspect_ratio
        self.max_aspect_ratio = max_aspect_ratio
        self.image = Image

    def process(self, file: "File", attr_key: str) -> None:
        super().process(file, attr_key)
        import PIL

        try:
            image = self.image.open(file.original_content)
        except (PIL.UnidentifiedImageError, OSError):
            raise InvalidImageError(attr_key, "Provide valid image file")
        width, height = image.width, image.height
        if self.min_width and width < self.min_width:
            raise DimensionValidationError(
                attr_key,
                f"Minimum allowed width is: {self.min_width}, but {width} is given.",
            )
        if self.min_height and height < self.min_height:
            raise DimensionValidationError(
                attr_key,
                f"Minimum allowed height is: {self.min_height}, but {height} is given.",
            )

        if self.max_width and self.max_width < width:
            raise DimensionValidationError(
                attr_key,
                f"Maximum allowed width is: {self.max_width}, but {width} is given.",
            )

        if self.max_height and self.max_height < height:
            raise DimensionValidationError(
                attr_key,
                f"Maximum allowed height is: {self.max_height}, but {height} is given.",
            )
        aspect_ratio = width / height
        if (self.min_aspect_ratio and self.min_aspect_ratio > aspect_ratio) or (
            self.max_aspect_ratio and self.max_aspect_ratio < aspect_ratio
        ):
            raise AspectRatioValidationError(
                attr_key,
                f"Invalid aspect ratio {width} / {height} = {aspect_ratio},"
                "accepted_range: "
                f"{self.min_aspect_ratio} - {self.max_aspect_ratio}",
            )
        file.update({"width": width, "height": height})
        file.original_content.seek(0)



import mimetypes
from abc import abstractmethod
from tempfile import SpooledTemporaryFile
from typing import TYPE_CHECKING, Optional, Tuple

from sqlalchemy_file.helpers import INMEMORY_FILESIZE

if TYPE_CHECKING:
    from sqlalchemy_file.file import File


class Processor:
    """
    Interface that must be implemented by file processors.
    Can be used to add additional data to the stored file or change it.
    When file processors are run the file has already been stored.

    """

    @abstractmethod
    def process(
        self, file: "File", upload_storage: Optional[str] = None
    ) -> None:  # pragma: no cover
        """
        Should be overridden in inherited class
        Parameters:
            file: [File][sqlalchemy_file.file.File] object,
                Use file.original_content to access uploaded file
            upload_storage: pass this to
                [file.store_content()][sqlalchemy_file.file.File.store_content]
                to attach additional files to the original file
        """
        pass


class ThumbnailGenerator(Processor):
    """
    Generate thumbnail from original content.

    The default thumbnail format and size are `PNG@128x128`, those can be changed
    by giving custom `thumbnail_size` and `thumbnail_format`

    !!! note
        ThumbnailGenerator will add additional data
        to the file object under the key `thumbnail`.
        These data will be store in database.

        Properties available in `thumbnail` attribute

        - **file_id:** This is the ID of the uploaded thumbnail file
        - **path:**    This is a upload_storage/file_id path which can
                       be used with :meth:`StorageManager.get_file` to
                       retrieve the thumbnail file
        - **width**    This is the width of the thumbnail image
        - **height:**  This is the height of the thumbnail image
        - **url:**     Public url of the uploaded file provided
                       by libcloud method `Object.get_cdn_url()`

    Example:
        ```Python
        class Book(Base):
            __tablename__ = "book"

            id = Column(Integer, autoincrement=True, primary_key=True)
            title = Column(String(100), unique=True)
            cover = Column(ImageField(processors=[ThumbnailGenerator()]))
        ```

        ```Python
        def test_create_image_with_thumbnail(self, fake_image) -> None:
            with Session(engine) as session:
                from PIL import Image

                session.add(Book(title="Pointless Meetings", cover=fake_image))
                session.flush()
                book = session.execute(
                    select(Book).where(Book.title == "Pointless Meetings")
                ).scalar_one()
                assert book.cover["thumbnail"] is not None
                thumbnail = StorageManager.get_file(book.cover["thumbnail"]["path"])
                assert thumbnail is not None
                thumbnail = Image.open(thumbnail)
                assert max(thumbnail.width, thumbnail.height) == 128
                assert book.cover["thumbnail"]["width"] == thumbnail.width
                assert book.cover["thumbnail"]["height"] == thumbnail.height
        ```


    """

    def __init__(
        self,
        thumbnail_size: Tuple[int, int] = (128, 128),
        thumbnail_format: str = "PNG",
    ) -> None:
        super().__init__()
        self.thumbnail_size = thumbnail_size
        self.thumbnail_format = thumbnail_format

    def process(self, file: "File", upload_storage: Optional[str] = None) -> None:
        from PIL import Image

        content = file.original_content
        img = Image.open(content)
        thumbnail = img.copy()
        thumbnail.thumbnail(self.thumbnail_size)
        output = SpooledTemporaryFile(INMEMORY_FILESIZE)
        thumbnail.save(output, self.thumbnail_format)
        output.seek(0)
        width, height, content_type = (
            thumbnail.width,
            thumbnail.height,
            f"image/{self.thumbnail_format}".lower(),
        )
        ext = mimetypes.guess_extension(content_type)
        stored_file = file.store_content(
            output,
            upload_storage,
            metadata={
                "filename": file["filename"] + f".thumbnail{width}x{height}{ext}",
                "content_type": content_type,
                "width": width,
                "height": height,
            },
        )
        file.update(
            {
                "thumbnail": {
                    "file_id": stored_file.name,
                    "width": width,
                    "height": height,
                    "path": "%s/%s" % (upload_storage, stored_file.name),
                    "url": stored_file.get_cdn_url(),
                }
            }
        )



from typing import Any, Dict, List, Optional, Set, Tuple, Type, Union

from sqlalchemy import event, inspect, orm, types
from sqlalchemy.engine import Connection, Dialect
from sqlalchemy.orm import ColumnProperty, Mapper, Session, SessionTransaction
from sqlalchemy.orm.attributes import get_history
from sqlalchemy_file.file import File
from sqlalchemy_file.mutable_list import MutableList
from sqlalchemy_file.processors import Processor, ThumbnailGenerator
from sqlalchemy_file.storage import StorageManager
from sqlalchemy_file.validators import ImageValidator, Validator


class FileField(types.TypeDecorator):  # type: ignore

    """
    Provides support for storing attachments to **SQLAlchemy** models.

    [FileField][sqlalchemy_file.types.FileField] can be used as a Column type to
    store files into the model. The actual file itself will be uploaded to a specific
    `libcloud.storage.base.Container`, and only the [File][sqlalchemy_file.file.File]
    information will be stored on the database as JSON.

    [FileField][sqlalchemy_file.types.FileField] is transaction aware, so it will delete
    every uploaded file whenever the transaction is rolled back and will
    delete any old file whenever the transaction is committed.

    You can save `str`, `bytes` or any python `file` object

    Each file will be validated by provided validators before being saved into
    associate storage `libcloud.storage.base.Container` and can go through different
    processors before being saved in the database.

    """

    impl = types.JSON
    cache_ok = True

    def __init__(
        self,
        *args: Tuple[Any],
        upload_storage: Optional[str] = None,
        validators: Optional[List[Validator]] = None,
        processors: Optional[List[Processor]] = None,
        upload_type: Type[File] = File,
        multiple: Optional[bool] = False,
        **kwargs: Dict[str, Any],
    ) -> None:
        """
        Parameters:
              upload_storage: storage to use
              validators: List of validators to apply
              processors: List of validators to apply
              upload_type: File class to use, could be
                        use to set custom File class
              multiple: Use this to save multiple files
        """
        super().__init__(*args, **kwargs)
        if processors is None:
            processors = []
        if validators is None:
            validators = []
        self.upload_storage = upload_storage
        self.upload_type = upload_type
        self.multiple = multiple
        self.validators = validators
        self.processors = processors

    def process_bind_param(
        self, value: Any, dialect: Dialect
    ) -> Union[None, Dict[str, Any], List[Dict[str, Any]]]:
        if not value:
            return None
        if not self.multiple and not isinstance(
            value, self.upload_type
        ):  # pragma: no cover
            raise ValueError(f"Expected {self.upload_type}, received: {type(value)}")
        if self.multiple and not (
            isinstance(value, list)
            and all([isinstance(v, self.upload_type) for v in value])
        ):  # pragma: no cover
            raise ValueError(
                f"Expected MutableList[{self.upload_type}], received: {type(value)}"
            )
        return [v.encode() for v in value] if self.multiple else value.encode()

    def process_result_value(
        self, value: Any, dialect: Dialect
    ) -> Union[None, MutableList[File], File]:
        if value is None:
            return None
        if type(value) is dict:
            return (
                MutableList([self.upload_type.decode(value)])
                if self.multiple
                else self.upload_type.decode(value)
            )
        return MutableList([self.upload_type.decode(v) for v in value])


class ImageField(FileField):
    """Inherits all attributes and methods from [FileField][sqlalchemy_file.types.FileField],
    but also validates that the uploaded object is a valid image.
    """

    def __init__(
        self,
        *args: Tuple[Any],
        upload_storage: Optional[str] = None,
        thumbnail_size: Optional[Tuple[int, int]] = None,
        image_validator: Optional[ImageValidator] = None,
        validators: Optional[List[Validator]] = None,
        processors: Optional[List[Processor]] = None,
        upload_type: Type[File] = File,
        multiple: Optional[bool] = False,
        **kwargs: Dict[str, Any],
    ) -> None:
        """
        Parameters:
            upload_storage: storage to use
            image_validator: ImageField use default image
                  validator, Use this property to customize it.
            thumbnail_size: If set, a thumbnail will be generated
                from original image using [ThumbnailGenerator]
                [sqlalchemy_file.processors.ThumbnailGenerator]
            validators: List of additional validators to apply
            processors: List of validators to apply
            upload_type: File class to use, could be
                        use to set custom File class
            multiple: Use this to save multiple files
        """
        if validators is None:
            validators = []
        if image_validator is None:
            image_validator = ImageValidator()
        if thumbnail_size is not None:
            if processors is None:
                processors = []
            processors.append(ThumbnailGenerator(thumbnail_size))
        validators.append(image_validator)
        super().__init__(
            *args,
            upload_storage=upload_storage,
            validators=validators,
            processors=processors,
            upload_type=upload_type,
            multiple=multiple,
            **kwargs,
        )


class FileFieldSessionTracker(object):
    mapped_entities: Dict[Type[Any], List[str]] = dict()

    @classmethod
    def delete_files(cls, paths: Set[str], ctx: str) -> None:
        for path in paths:
            StorageManager.delete_file(path)

    @classmethod
    def clear_session(cls, session: Session) -> None:
        if hasattr(session, "_new_files"):
            del session._new_files  # type: ignore
        if hasattr(session, "_old_files"):
            del session._old_files  # type: ignore

    @classmethod
    def add_new_files_to_session(cls, session: Session, paths: List[str]) -> None:
        session._new_files = getattr(session, "_new_files", set())  # type: ignore
        session._new_files.update(paths)  # type: ignore

    @classmethod
    def add_old_files_to_session(cls, session: Session, paths: List[str]) -> None:
        session._old_files = getattr(session, "_old_files", set())  # type: ignore
        session._old_files.update(paths)  # type: ignore

    @classmethod
    def extract_files_from_history(
        cls, data: List[Union[MutableList[File], File]]
    ) -> List[str]:
        paths = []
        for item in data:
            if isinstance(item, list):
                paths.extend([f["path"] for f in item])
            elif isinstance(item, File):
                paths.append(item["path"])
        return paths

    @classmethod
    def _mapper_configured(cls, mapper: Mapper, class_: Any) -> None:
        """Detect and listen all class with FileField Column"""
        for mapper_property in mapper.iterate_properties:
            if isinstance(mapper_property, ColumnProperty) and isinstance(
                mapper_property.columns[0].type, FileField
            ):
                assert (
                    len(mapper_property.columns) == 1
                ), "Multiple-column properties are not supported"
                if mapper_property.columns[0].type.multiple:
                    MutableList.associate_with_attribute(
                        getattr(class_, mapper_property.key)
                    )
                cls.mapped_entities.setdefault(class_, []).append(mapper_property.key)

    @classmethod
    def _after_configured(cls) -> None:
        for entity in cls.mapped_entities.keys():
            event.listen(entity, "before_insert", cls._before_insert)
            event.listen(entity, "before_update", cls._before_update)
            event.listen(entity, "after_update", cls._after_update)
            event.listen(entity, "after_delete", cls._after_delete)

    @classmethod
    def _after_commit(cls, session: Session) -> None:
        """After commit, old files are automatically deleted"""
        cls.delete_files(getattr(session, "_old_files", set()), "after_commit")
        cls.clear_session(session)

    @classmethod
    def _after_soft_rollback(cls, session: Session, _: SessionTransaction) -> None:
        """After rollback, new files are automatically deleted"""
        cls.delete_files(getattr(session, "_new_files", set()), "after_soft_rollback")
        cls.clear_session(session)

    @classmethod
    def _after_delete(cls, mapper: Mapper, _: Connection, obj: Any) -> None:
        """
        After delete mark all linked files as old in order to delete
        them when after session is committed
        """
        tracked_columns: List[str] = cls.mapped_entities.get(mapper.class_, [])
        for key in tracked_columns:
            value = getattr(obj, key)
            if value is not None:
                cls.add_old_files_to_session(
                    inspect(obj).session,
                    [
                        f["path"]
                        for f in (value if isinstance(value, list) else [value])
                    ],
                )

    @classmethod
    def _after_update(cls, mapper: Mapper, _: Connection, obj: Any) -> None:
        """
        After update, mark all edited files as old
        in order to delete them when after session is committed
        """
        tracked_columns: List[str] = cls.mapped_entities.get(mapper.class_, [])
        for key in tracked_columns:
            history = get_history(obj, key)
            cls.add_old_files_to_session(
                inspect(obj).session, cls.extract_files_from_history(history.deleted)
            )

    @classmethod
    def _before_update(cls, mapper: Mapper, _: Connection, obj: Any) -> None:
        """
        Before updating values, validate and save files. For multiple fields,
        mark all removed files as old, as _removed attribute will be
        reinitialised after update.
        """
        session = inspect(obj).session
        tracked_columns: List[str] = cls.mapped_entities.get(mapper.class_, [])
        for key in tracked_columns:
            value = getattr(obj, key)
            if value is not None:
                changed, prepare_value = cls.prepare_file_attr(mapper, obj, key)
                if changed:
                    setattr(obj, key, prepare_value)
                    history = get_history(obj, key)
                    cls.add_new_files_to_session(
                        session, cls.extract_files_from_history(history.added)
                    )
                if isinstance(value, MutableList):
                    _removed = getattr(value, "_removed", ())
                    cls.add_old_files_to_session(session, [f["path"] for f in _removed])

    @classmethod
    def _before_insert(cls, mapper: Mapper, _: Connection, obj: Any) -> None:
        """Before inserting values, mark all created files as new. They will be
        automatically removed when session rollback"""

        tracked_columns: List[str] = cls.mapped_entities.get(mapper.class_, [])
        for key in tracked_columns:
            value = getattr(obj, key)
            if value is not None:
                setattr(obj, key, cls.prepare_file_attr(mapper, obj, key)[1])
                history = get_history(obj, key)
                cls.add_new_files_to_session(
                    inspect(obj).session, cls.extract_files_from_history(history.added)
                )

    @classmethod
    def prepare_file_attr(
        cls, mapper: Mapper, obj: Any, key: str
    ) -> Tuple[bool, Union[File, List[File]]]:
        """
        Prepare file before saved to database, convert bytes and string,
        saved file into the upload_storage, apply validators and processors
        """
        value = getattr(obj, key)

        """Become True when it is new file for single field,
        or when new items is added for multiple field"""
        changed = False

        column_type = mapper.attrs.get(key).columns[0].type
        assert isinstance(column_type, FileField)
        upload_type = column_type.upload_type

        prepared_values: List[File] = []
        for v in value if isinstance(value, list) else [value]:
            if not isinstance(v, upload_type):
                v = upload_type(v)
            if not getattr(v, "saved", False):
                changed = True
                v.apply_validators(column_type.validators, key)
            prepared_values.append(v)

        upload_storage = column_type.upload_storage or StorageManager.get_default()
        for value in prepared_values:
            if not getattr(value, "saved", False):
                value.save_to_storage(upload_storage)
                value.apply_processors(column_type.processors, upload_storage)
        return changed, (
            prepared_values if column_type.multiple else prepared_values[0]
        )

    @classmethod
    def setup(cls) -> None:
        event.listen(orm.mapper, "mapper_configured", cls._mapper_configured)
        event.listen(orm.mapper, "after_configured", cls._after_configured)
        event.listen(Session, "after_commit", cls._after_commit)
        event.listen(Session, "after_soft_rollback", cls._after_soft_rollback)


FileFieldSessionTracker.setup()



import typing
from typing import Any


class BaseFile(typing.Dict[str, Any]):
    """
    Base class for file object. It keeps information on a content related to a specific storage.
    It is a specialized dictionary that provides also attribute style access,
    the dictionary parent permits easy encoding/decoding to JSON.

    """

    def __getitem__(self, key: str) -> Any:
        return dict.__getitem__(self, key)

    def __getattr__(self, name: str) -> Any:
        try:
            return self[name]
        except KeyError:
            raise AttributeError(name)

    def __setitem__(self, key: str, value: Any) -> None:
        if getattr(self, "_frozen", False):
            raise TypeError("Already saved files are immutable")
        return dict.__setitem__(self, key, value)

    __setattr__ = __setitem__

    def __delattr__(self, name: str) -> None:
        if getattr(self, "_frozen", False):
            raise TypeError("Already saved files are immutable")

        try:
            del self[name]
        except KeyError:
            raise AttributeError(name)

    def __delitem__(self, key: str) -> None:
        if object.__getattribute__(self, "_frozen"):
            raise TypeError("Already saved files are immutable")
        dict.__delitem__(self, key)

    def _freeze(self) -> None:
        object.__setattr__(self, "_frozen", True)

    def _thaw(self) -> None:
        object.__setattr__(self, "_frozen", False)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['sqlalchemy_file']

package_data = \
{'': ['*']}

install_requires = \
['SQLAlchemy>=1.4,<1.5.0', 'apache-libcloud>=3.6.0,<4.0.0']

setup_kwargs = {
    'name': 'sqlalchemy-file',
    'version': '0.1.3',
    'description': 'SQLAlchemy-file is a SQLAlchemy extension for attaching files to SQLAlchemy model and uploading them to various storage.',
    'long_description': '# sqlalchemy-file\n\n\n**SQLAlchemy-file** is a [SQLAlchemy](https://www.sqlalchemy.org/) extension for attaching files to SQLAlchemy model and\nuploading them to various storage such as Local Storage, Amazon S3, Rackspace CloudFiles, Google Storage and others\nusing [Apache Libcloud](https://github.com/apache/libcloud).\n\n<p align="center">\n<a href="https://github.com/jowilf/sqlalchemy-file/actions">\n    <img src="https://github.com/jowilf/sqlalchemy-file/actions/workflows/test.yml/badge.svg" alt="Test suite">\n</a>\n<a href="https://github.com/jowilf/sqlalchemy-file/actions">\n    <img src="https://github.com/jowilf/sqlalchemy-file/actions/workflows/publish.yml/badge.svg" alt="Publish">\n</a>\n<a href="https://codecov.io/gh/jowilf/sqlalchemy-file">\n    <img src="https://codecov.io/gh/jowilf/sqlalchemy-file/branch/main/graph/badge.svg" alt="Codecov">\n</a>\n<a href="https://pypi.org/project/sqlalchemy-file/">\n    <img src="https://badge.fury.io/py/sqlalchemy-file.svg" alt="Package version">\n</a>\n<a href="https://pypi.org/project/sqlalchemy-file/">\n    <img src="https://img.shields.io/pypi/pyversions/sqlalchemy-file?color=2334D058" alt="Supported Python versions">\n</a>\n</p>\n\n\nThe key features are:\n\n* **Multiple Storage :** Use Object Storage API provided by [Apache Libcloud](https://github.com/apache/libcloud) to\n  store files. Therefore, you can store your files on Local Storage, Amazon S3, Google Cloud Storage, MinIO etc, and\n  easily switch between them. For a full list of supported providers\n  visit [supported providers page](https://libcloud.readthedocs.io/en/stable/storage/supported_providers.html) from Apache\n  Libcloud documentation.\n* **Validator :**  Provide an interface for validating each files before saving them.\n* **Size Validator :** Built-in validator for file maximum `size` validation.\n* **Content-Type Validator :** Built-in validator for file ``mimetype`` restrictions.\n* **Image Validator :** Built-in validator for image `mimetype`, `width`, `height` and `ratio` validation.\n* **Processor :** Provide an interface to easily save multiple transformation of the original files.\n* **ThumbnailGenerator :** Built-in processor to auto generate thumbnail\n* **Multiple Files :** You can attach multiple files directly to a Model.\n* **Session awareness :** Whenever an object is deleted or a rollback is performed the files uploaded during the unit of\n  work or attached to the deleted objects are automatically deleted.\n* **Meant for Evolution :** Change the storage provider anytime you want, old data will continue to work\n* **SQLModel Support:** Tested with [SQLModel](https://github.com/tiangolo/sqlmodel)\n\n---\n\n**Documentation**: [https://jowilf.github.io/sqlalchemy-file](https://jowilf.github.io/sqlalchemy-file/)\n\n**Source Code**: [https://github.com/jowilf/sqlalchemy-file](https://github.com/jowilf/sqlalchemy-file)\n\n---\n\n## Requirements\n\nA recent and currently supported version of Python (right\nnow, <a href="https://www.python.org/downloads/" class="external-link" target="_blank">Python supports versions 3.7 and\nabove</a>).\n\nAs **SQLAlchemy-file** is based on **Apache Libcloud** and **SQLAlchemy**, it requires them. They will be automatically\ninstalled when you install SQLAlchemy-file.\n\n## Installation\n\n### PIP\n\n```shell\n$ pip install sqlalchemy-file\n```\n\n### Poetry\n\n```shell\n$ poetry add sqlalchemy-file\n```\n\n## Example\n\nAttaching files to models is as simple as declaring a field on the model itself\n\n```Python\nimport os\n\nfrom libcloud.storage.drivers.local import LocalStorageDriver\nfrom sqlalchemy import Column, Integer, String, create_engine\nfrom sqlalchemy.ext.declarative import declarative_base\nfrom sqlalchemy.orm import Session\nfrom sqlalchemy_file import File, FileField\nfrom sqlalchemy_file.storage import StorageManager\n\nBase = declarative_base()\n\n\n# Define your model\nclass Attachment(Base):\n    __tablename__ = "attachment"\n\n    id = Column(Integer, autoincrement=True, primary_key=True)\n    name = Column(String(50), unique=True)\n    content = Column(FileField)\n\n\n# Configure Storage\nos.makedirs("/tmp/storage/attachment", 0o777, exist_ok=True)\ncontainer = LocalStorageDriver("/tmp/storage").get_container("attachment")\nStorageManager.add_storage("default", container)\n\n# Save your model\nengine = create_engine(\n    "sqlite:///example.db", connect_args={"check_same_thread": False}\n)\nBase.metadata.create_all(engine)\n\nwith Session(engine) as session:\n    session.add(Attachment(name="attachment1", content=open("./example.txt", "rb")))\n    session.add(Attachment(name="attachment2", content=b"Hello world"))\n    session.add(Attachment(name="attachment3", content="Hello world"))\n    file = File(content="Hello World", filename="hello.txt", content_type="text/plain")\n    session.add(Attachment(name="attachment4", content=file))\n    session.commit()\n\n```\n\n## Related projects and inspirations\n\n* [filedepot: ](https://github.com/amol-/depot) When I was looking for a library like this, depot was the\nbest I saw. This project inspired **SQLAlchemy-file** extensively\nand some features are implemented the same.\n* [sqlalchemy-media](https://github.com/pylover/sqlalchemy-media) Another attachment extension for SqlAlchemy \nto manage assets which are associated with database models\n\n',
    'author': 'Jocelin Hounon',
    'author_email': 'hounonj@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://jowilf.github.io/sqlalchemy-file',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from typing import Any, Dict, Iterator, Optional

from libcloud.storage.base import Container
from libcloud.storage.types import ObjectDoesNotExistError
from sqlalchemy_file.helpers import LOCAL_STORAGE_DRIVER_NAME, get_metadata_file_obj
from sqlalchemy_file.stored_file import StoredFile


class StorageManager:
    """
    Takes care of managing the whole Storage environment for the application.

    Use [add_storage][sqlalchemy_file.storage.StorageManager.add_storage] method
    to add new `libcloud.storage.base.Container`and associate a name which
    will be use later to retrieve this container.

    The first container will be used as default, to simplify code when you have
    only one container.

    Use associated name as `upload_storage` for [FileField][sqlalchemy_file.types.FileField]
    to store his files inside the corresponding container.

    """

    _default_storage_name: Optional[str] = None
    _storages: Dict[str, Container] = {}

    @classmethod
    def set_default(cls, name: str) -> None:
        """Replaces the current application default storage"""
        if name not in cls._storages:
            raise RuntimeError("%s storage has not been added" % (name,))
        cls._default_storage_name = name

    @classmethod
    def get_default(cls) -> str:
        """Gets the current application default storage"""
        if cls._default_storage_name is None:
            raise RuntimeError("No default storage has been added")
        return cls._default_storage_name

    @classmethod
    def add_storage(cls, name: str, container: Container) -> None:
        """Add new storage"""
        assert isinstance(container, Container), "Invalid container"
        if name in cls._storages:
            raise RuntimeError("Storage %s has already been added" % (name,))
        if cls._default_storage_name is None:
            cls._default_storage_name = name
        cls._storages[name] = container

    @classmethod
    def get(cls, name: Optional[str] = None) -> Container:
        """
        Gets the container instance associate to the name,
        return default if name isn't provided
        """
        if name is None and cls._default_storage_name is None:
            raise RuntimeError("No default storage have been added")
        elif name is None:
            name = cls._default_storage_name
        if name in cls._storages:
            return cls._storages[name]
        raise RuntimeError("%s storage has not been added" % (name,))

    @classmethod
    def save_file(
        cls,
        name: str,
        content: Iterator[bytes],
        upload_storage: Optional[str] = None,
        metadata: Optional[Dict[str, Any]] = None,
        headers: Optional[Dict[str, str]] = None,
    ) -> StoredFile:
        """Save file into provided `upload_storage`"""
        container = cls.get(upload_storage)
        if container.driver.name == LOCAL_STORAGE_DRIVER_NAME:
            obj = container.upload_object_via_stream(iterator=content, object_name=name)
            if metadata is not None:
                """
                Libcloud local storage driver doesn't support metadata, so the metadata
                is saved in the same container with the combination of the original name
                and `.metadata.json` as name
                """
                container.upload_object_via_stream(
                    iterator=get_metadata_file_obj(metadata),
                    object_name=f"{name}.metadata.json",
                )
            return StoredFile(obj)
        else:
            extra = {}
            if metadata is not None:
                if "content_type" in metadata:
                    extra["content_type"] = metadata["content_type"]
                extra["meta_data"] = metadata
            return StoredFile(
                container.upload_object_via_stream(
                    iterator=content, object_name=name, extra=extra, headers=headers
                )
            )

    @classmethod
    def get_file(cls, path: str) -> StoredFile:
        """Retrieve the file with `provided` path
        path is expected to be `storage_name/file_id`
        """
        upload_storage, file_id = path.split("/")
        return StoredFile(StorageManager.get(upload_storage).get_object(file_id))

    @classmethod
    def delete_file(cls, path: str) -> bool:
        """Delete the file with `provided` path
        path is expected to be `storage_name/file_id`
        """
        upload_storage, file_id = path.split("/")
        obj = StorageManager.get(upload_storage).get_object(file_id)
        if obj.driver.name == LOCAL_STORAGE_DRIVER_NAME:
            """Try deleting associated metadata file"""
            try:
                obj.container.get_object(f"{obj.name}.metadata.json").delete()
            except ObjectDoesNotExistError:
                pass
        return obj.delete()

    @classmethod
    def _clear(cls) -> None:
        """
        This is only for testing pourposes, resets the StorageManager
        """
        cls._default_storage_name = None
        cls._storages = {}
